import torch
import torch.nn as nn
import torch.optim as optim
from sklearn.metrics import classification_report, confusion_matrix
from torch.autograd import Variable
from torch.utils.data import DataLoader, TensorDataset
from sklearn.metrics import roc_curve, auc
import matplotlib.pyplot as plt
import matplotlib as mpl
from torch.optim.lr_scheduler import StepLR
import pandas as pd


# =============数据集导入==============
# 读取 Excel 文件
df1 = pd.read_excel('G:\Data-master\GM_gasf_python169均值填补.xlsx')
df2 = pd.read_excel('G:\Data-master\GM_gadf_python169均值填补.xlsx')
RP3 = pd.read_excel('G:\Data-master\RP矩阵——NA均值填补dim_3.xlsx')  #121个RP特征 特征1
print(df1)
print(df2)
print(RP3)

gasf_df = df1[['63', '50', '120', '84', '56', '67', '14', '112', '4', '2']].copy()
gasf_df.columns = ['sf63', 'sf50', 'sf120', 'sf84', 'sf56', 'sf67', 'sf14', 'sf112', 'sf4', 'sf2']
print(gasf_df)

gadf_df = df2[['2', '10',  '97', '120', '93', '162']].copy()
gadf_df.columns = ['df2', 'df10', 'df97', 'df120', 'df93', 'df162']

rp_df = RP3[['1']].copy()
rp_df.columns = ['3rp1']
event = df2[['id', 'DFS']].copy()


T13data = pd.read_excel(r'G:\Data-master\EBVfeture752_T3-T13_diflength\EBVfeature_T13_diflength752.xlsx')
baseline_data = T13data[['sex', 'age_45', 'smoke', 'T', 'N', 'stage']]
#SF_DF_RP = pd.concat([gasf_df, gadf_df, rp_df, event], axis=1)
data = pd.concat([gasf_df, gadf_df, rp_df, baseline_data, event], axis=1)



def data_partition(all_df, excel_file):
    # 根据所提供的ID将 'id' 列设为索引
    all_df.set_index('id', inplace=True)

    # 从Excel文件中读取训练集ID
    train_ids = pd.read_excel(excel_file).iloc[:, 0]
    # 使用这些ID从final_data中获取训练集
    train_data = all_df.loc[train_ids]
    # 剩下的数据作为测试集
    test_data = all_df.drop(train_ids)

    # X提取变量特征；Y提取目标变量
    X_train = train_data.drop(['DFS'], axis=1)
    y_train = train_data['DFS'].values
    X_test = test_data.drop(['DFS'], axis=1)
    y_test = test_data['DFS'].values

    feature = list(X_train.columns)
    X_train = X_train.values
    X_test = X_test.values
    print("训练集维度：", train_data.shape)
    print("测试集维度：", test_data.shape)
    print("test_data的索引：", test_data.columns)

    # 将numpy数组转换为torch tensors
    X_train = Variable(torch.Tensor(X_train).float())
    y_train = Variable(torch.Tensor(y_train).long())
    X_test = Variable(torch.Tensor(X_test).float())
    y_test = Variable(torch.Tensor(y_test).long())

    return X_train, y_train, X_test, y_test, feature



class MLP(nn.Module):
    def __init__(self):
        super(MLP, self).__init__()

        # 定义输入层到隐藏层的连接，输入数据维度为752*24，隐藏层宽度为100
        self.layer1 = nn.Linear(23, 100)

        # 定义隐藏层到隐藏层的连接，隐藏层宽度从100变为200
        self.layer2 = nn.Linear(100, 200)

        # 定义隐藏层到隐藏层的连接，隐藏层宽度从200变为300
        self.layer3 = nn.Linear(200, 300)

        # 定义输出层，从最后一个隐藏层到输出，输出层的结点数为2，代表二分类
        self.output = nn.Linear(300, 2)

        # 定义ReLU激活函数
        self.relu = nn.ReLU()

    def forward(self, x):
        # 输入层到隐藏层的操作
        x = self.layer1(x)
        x = self.relu(x)

        # 隐藏层到隐藏层的操作
        x = self.layer2(x)
        x = self.relu(x)

        # 隐藏层到隐藏层的操作
        x = self.layer3(x)
        x = self.relu(x)

        # 输出层的操作
        x = self.output(x)

        return x



def train(model, loader, criterion, optimizer):
    model.train()
    running_loss = 0.0
    corrects = 0
    for i, data in enumerate(loader, 0):
        inputs, labels = data

        if torch.cuda.is_available():
            inputs, labels = Variable(inputs.cuda()), Variable(labels.cuda())
        else:
            inputs, labels = Variable(inputs), Variable(labels)

        outputs = model(inputs)
        _, preds = torch.max(outputs.data, 1)
        optimizer.zero_grad()
        loss = criterion(outputs, labels)
        loss.backward()
        optimizer.step()
        running_loss += loss.item()
        corrects += torch.sum(preds == labels.data)

    epoch_loss = running_loss / len(loader.dataset)
    epoch_acc = corrects.double() / len(loader.dataset)
    print('Train Loss: {:.4f} Acc: {:.4f}'.format(epoch_loss, epoch_acc))
    return epoch_loss

def evaluate(model, loader):
    model.eval()
    running_loss = 0.0
    corrects = 0
    predictions = []  # 保存模型的预测得分
    truths = []  # 保存实际标签
    criterion = nn.CrossEntropyLoss()  # 定义计算损失的函数
    with torch.no_grad():
        for i, data in enumerate(loader):
            inputs, labels = data
            if torch.cuda.is_available():
                inputs, labels = inputs.to('cuda'), labels.to('cuda')
            outputs = model(inputs)
            _, preds = torch.max(outputs.data, 1)
            loss = criterion(outputs, labels)
            running_loss += loss.item()
            corrects += torch.sum(preds == labels.data)
            predictions.extend(torch.nn.functional.softmax(outputs.data, dim=1)[:,
                               1].cpu().numpy())  # extacting positive class prediction probabilities
            #将模型的预测输出从对数概率转变为实际的概率，并选择了正类的概率作为预测得分
            truths.extend(labels.data.cpu().numpy())  # 需要将实际标签转为numpy格式

    epoch_loss = running_loss / len(loader)
    epoch_acc = corrects.double() / len(loader.dataset)
    print('Val Loss: {:.4f} Acc: {:.4f}'.format(epoch_loss, epoch_acc))
    # 将概率预测转变为类别预测
    predictions_binary = [1 if p >= 0.5 else 0 for p in predictions]

    print(classification_report(truths, predictions_binary))
    # 然后，就可以使用confusion_matrix函数
    print(confusion_matrix(truths, predictions_binary))
    #print(classification_report(truths, predictions))
    #print(confusion_matrix(truths, predictions))
    return predictions_binary, truths, epoch_loss



def MLP_pytorch(X_train, y_train, X_test, y_test):
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    model = MLP().to(device)
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(model.parameters(), lr=0.01)  # 设置初始学习率
    scheduler = StepLR(optimizer, step_size=10, gamma=0.1)  # 每10个epochs，学习率乘以0.1
    epochs = 100
    batch = 30
    train_losses = []
    val_losses = []

    train_ds = TensorDataset(torch.Tensor(X_train), torch.Tensor(y_train))
    train_loader = DataLoader(train_ds, batch_size=batch, shuffle=True)
    test_ds = TensorDataset(torch.Tensor(X_test), torch.Tensor(y_test))
    test_loader = DataLoader(test_ds, batch_size=batch, shuffle=False)

    for epoch in range(epochs):
        print('Epoch {}/{}'.format(epoch + 1, epochs))
        train(model, train_loader, criterion, optimizer)
        train_loss = train(model, train_loader, criterion, optimizer)
        scheduler.step()  # 在每个epoch之后调用step方法更新学习率
        predictions_binary, truths, val_loss = evaluate(model, test_loader)

        train_losses.append(train_loss)
        val_losses.append(val_loss)
    return predictions_binary, truths, train_losses, val_losses


        # 用PyTorch来训练了一个多层感知机(MLP)
    # 模型，使用了交叉熵损失(
    #     CrossEntropyLoss)，优化器选用的是Adam。模型训练完后，在测试集上验证了模型的精度，并打印了模型的分类报告和混淆矩阵。
    # 注意的是，在pytorch中我们一般使用DataLoader来进行数据的读取，这样可以很方便地实现批处理和数据打乱。
    # 但要确保你的数据X_train, y_train, X_test, y_test已经是numpy
    # array或者tensor格式

def Draw_ROC(true_labels, pred_scores):
    fpr, tpr, _ = roc_curve(true_labels, pred_scores)
    roc_auc = auc(fpr, tpr)

    plt.plot(fpr, tpr, 'b', label='AUC = %0.2f' % roc_auc)
    plt.legend(loc='lower right', fontsize=12)
    plt.plot([0, 1], [0, 1], 'r--')
    plt.ylabel('True Positive Rate', fontsize=12)
    plt.xlabel('False Positive Rate', fontsize=12)
    plt.title('Receiver Operating Characteristic (ROC)')
    plt.figure()
    plt.show()

def plot_losses(train_losses, val_losses):
    plt.plot(train_losses, label='Training loss')
    plt.plot(val_losses, label='Validation loss')
    plt.legend()
    plt.ylabel('Loss')
    plt.xlabel('epoch')
    plt.figure()
    plt.show()



if __name__ == "__main__":
    excel_file = r'G:\Data-master\EBVfeture752_T3-T13_diflength\seed 4264 difleng\EBVfeatureT13_train.xlsx'

    # 划分数据集
    X_train, y_train, X_test, y_test, feature = data_partition(data, excel_file)
    predictions_binary, truths, train_losses, val_losses = MLP_pytorch(X_train, y_train, X_test, y_test)
    Draw_ROC(truths, predictions_binary)
    plot_losses(train_losses, val_losses)






